local MAJOR = "StarLibLCDText-1.0" 
local MINOR = 1
assert(LibStub, MAJOR.." requires LibStub") 
local LCDText = LibStub:NewLibrary(MAJOR, MINOR)
if not LCDText then return end
local core = LibStub("StarLibCore-1.0")
assert(core, MAJOR .. " requires StarLibCore-1.0")
local error = LibStub("StarLibError-1.0")
assert(error, MAJOR .. " requires StarLibError-1.0")
local LibBuffer = LibStub("StarLibBuffer-1.0", true)
assert(LibBuffer, MAJOR .. " requires StarLibBuffer-1.0")
local LibUtils = LibStub("StarLibUtils-1.0", true)
assert(LibUtils, MAJOR .. " requires StarLibUtils-1.0")

local pool = setmetatable({}, {__mode = "k"})

local objects = {}
local objectsDict = {}
local update
local frame = CreateFrame("Frame")

local DEFAULT_LIMIT = 50

if not LCDText.__index then
	LCDText.__index = LCDText
end

-- @name StarLCDText:New
-- @usage StarLCDText:New(visitor, rows, cols, yres, xres, layers)
-- @param visitor A StarLibCore object
-- @param rows The number of rows
-- @param cols The number of cols
-- @param yres The y resolution for characters
-- @param xres The x resolution for characters
-- @param layers The number of layers
-- @param errorLevel Error verbosity level
-- @return A new StarLCDText object

function LCDText:New(visitor, rows, cols, yres, xres, layers, errorLevel, realBlit, data)
	
	local obj = next(pool)

	if obj then
		pool[obj] = nil
	else
		obj = {}
	end

	setmetatable(obj, self)
	
	tinsert(objects, obj)
	objectsDict[obj] = obj

	obj.visitor = visitor
	obj.LROWS = rows
	obj.LCOLS = cols
	obj.DROWS = rows
	obj.DCOLS = cols
	obj.LAYERS = layers
	obj.errorLevel = errorLevel
	obj.RealBlit = realBlit
	obj.data = data

	local n = rows * cols
	
	obj.LayoutFB = {}
	obj.TransitionFB = {}
	for i = 0, layers do
		obj.LayoutFB[i] = LibBuffer:New(n, self.errorLEvel)
		obj.TransitionFB[i] = LibBuffer:New(n, self.errorLevel)
	end
	obj.DisplayFB = LibBuffer:New(n, self.errorLevel)
	
	return obj	
	
end

-- @name StarLCDText:Del
-- @usage StarLCDText:Del([lcd]) or object:Del()
-- @param lcd An optional lcd object
-- @return Nothing
function LCDText:Del(lcd)
	
	if lcd and type(lcd) == "table" then
		pool[lcd] = true
		for i, v in ipairs(objects) do
			if v == lcd then
				objectsDict[lcd] = nil
				tremove(objects, i)
			end
		end
	else
		pool[self] = true
		for i, v in ipairs(objects) do
			if v == lcd then
				objectsDict[lcd] = nil
				tremove(objects, i)
			end
		end
	end
end

function LCDText:TextBlit(row, col, height, width)
	local lr, lc
	local dr, dc
	local p1, p2
	local fb = LibBuffer:New(self.LROWS * self.LCOLS)
	
	local r = row
	--[[
	while r < self.LROWS and r < row + height do
		local c = col
		while c < self.LCOLS and c < col + width do
			local n = r * self.LCOLS + c
			local l = 0
			while l < self.LAYERS do
				if self.LayoutFB[l].buffer[n] ~= ' ' then
					fb:Replace(n, self.LayoutFB[l].buffer[n])
					
				end
				l = l + 1
			end
			c = c + 1
		end
		r = r + 1
	end
	]]
	fb:Copy(self.LayoutFB[0])

	--/* loop over layout rows */
	lr = row
	
	while lr < self.LROWS and lr < row + height do
		--/* transform layout to display row */
		dr = lr;
		--/* sanity check */
		if dr >= 0 and dr < self.DROWS then 		
			--/* loop over layout cols */
			lc = col
			while lc < self.LCOLS and lc < col + width do
				--/* transform layout to display column */
				dc = lc;
				--/* sanity check */
				if dc >= 0 and dc < self.DCOLS then
					--/* find start of difference */
					if self.DisplayFB.buffer[dr * self.DCOLS + dc] ~= fb.buffer[lr * self.LCOLS + lc] then
					--if string.sub(self.DisplayFB, dr * self.DCOLS + dc, 1) ~= string.sub(fb, lr * self.LCOLS + lc) then
						--/* find end of difference */
						p1 = dc
						p2 = p1
						while lc < self.LCOLS and lc < col + width do
							--/* transform layout to display column */
							dc = lc;
							--/* sanity check */
							if dc >= 0 and dc < self.DCOLS then
								if self.DisplayFB.buffer[dr * self.DCOLS + dc] ~= fb.buffer[lr * self.LCOLS + lc] then
									p2 = dc;
								end
							end
							lc  = lc + 1
						end

						--/* send to display */
						local from = fb:Sub(lr * self.LCOLS + p1)
						self.DisplayFB:Merge(from, dr * self.DCOLS + p1, p2 - p1 + 1)
						if (self.RealBlit) then
							local tmp = self.DisplayFB:Sub(dr * self.DCOLS + p1)
							self:RealBlit(self.data, dr, p1, tmp.buffer, p2 - p1 + 1)
							tmp:Del()
						end
						from:Del()
					end
				end
				lc = lc + 1
			end
		end
		lr = lr + 1
	end
	fb:Del()
end

function LCDText.TextDraw(widget, data)
	local lcd = widget.visitor.lcd
	local r = widget.widget.row
	local c = widget.widget.col
	local n  = r * lcd.LCOLS + c
	local tmp = lcd.LayoutFB
	local len = strlen(widget.buffer)
	local num = 1
	for i = n, n + len do
		lcd.LayoutFB[widget.widget.layer]:Replace(i, widget.buffer:sub(num, num))
		num = num + 1
	end
	lcd:TextBlit(r, c, 1, len)
end